//-----------------------------------------------------------------------------
// Copyright (c) 2007-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Text;
using Dhpoware;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace XNAParallaxNormalMapping
{
    /// <summary>
    /// This Windows XNA application demonstrates tangent space parallax normal
    /// mapping. An effect (.fx) file is provided that implements tangent space
    /// parallax normal mapping using a series of HLSL shaders. The lighting
    /// models implemented are: directional lighting, point lighting, and
    /// spotlight lighting.
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        /// <summary>
        /// A light. This light structure is the same as the one defined in
        /// the parallax_normal_mapping.fx file. The only difference is the
        /// LightType enum.
        /// </summary>
        private struct Light
        {
            public enum LightType
            {
                DirectionalLight,
                PointLight,
                SpotLight
            }

            public LightType Type;
            public Vector3 Direction;
            public Vector3 Position;
            public Color Ambient;
            public Color Diffuse;
            public Color Specular;
            public float SpotInnerConeRadians;
            public float SpotOuterConeRadians;
            public float Radius;
        }

        /// <summary>
        /// A material. This material structure is the same as the one defined
        /// in the parallax_normal_mapping.fx file. We use the Color type here
        /// instead of a four element floating point array.
        /// </summary>
        private struct Material
        {
            public Color Ambient;
            public Color Diffuse;
            public Color Emissive;
            public Color Specular;
            public float Shininess;
        }

        private const float CAMERA_FOVY = 45.0f;
        private const float CAMERA_ZNEAR = 0.01f;
        private const float CAMERA_ZFAR = 100.0f;
        private const float CUBE_DOLLY_MIN = -10.0f;
        private const float CUBE_DOLLY_MAX = 0.0f;
        private const float CUBE_SIZE = 2.0f;
        private const float CUBE_SPEED_DOLLYING = 0.02f;
        private const float CUBE_SPEED_ROTATING = 0.30f;
        private const float CUBE_SPEED_TRACKING = 0.01f;
        private const float LIGHT_RADIUS = 10.0f;

        private GraphicsDeviceManager graphics;
        private SpriteFont spriteFont;
        private SpriteBatch spriteBatch;
        private Effect effect;
        private Texture2D nullTexture;
        private Texture2D colorMap;
        private Texture2D normalMap;
        private Texture2D heightMap;
        private NormalMappedCube cube;
        private Vector2 scaleBias;
        private Vector2 fontPos;
        private Vector2 cubeRotation;
        private Vector3 cubePos;
        private Vector3 cameraPos;
        private Color globalAmbient;
        private Light light;
        private Material material;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private MouseState currentMouseState;
        private MouseState prevMouseState;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private bool disableColorMap;
        private bool displayHelp;
        private bool disableParallax;

        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Parallax Normal Mapping Demo";
            IsMouseVisible = true;
            IsFixedTimeStep = false;
        }

        protected override void Initialize()
        {
            base.Initialize();

            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Setup cube.
            cubePos = new Vector3(0.0f, 0.0f, 0.0f);
            cubeRotation = new Vector2(0.0f, 0.0f);
            cube = new NormalMappedCube(GraphicsDevice, CUBE_SIZE);

            // Position the camera.
            cameraPos = new Vector3(0.0f, 0.0f, CUBE_SIZE * 2.0f);

            // Initial position for text rendering.
            fontPos = new Vector2(1.0f, 1.0f);

            // Initialize light settings.
            // Directional lighting is the default.
            globalAmbient = new Color(new Vector4(0.2f, 0.2f, 0.2f, 1.0f));
            light.Type = Light.LightType.DirectionalLight;
            light.Direction = Vector3.Forward;
            light.Position = cameraPos;
            light.Ambient = Color.White;
            light.Diffuse = Color.White;
            light.Specular = Color.White;
            light.SpotInnerConeRadians = MathHelper.ToRadians(20.0f);
            light.SpotOuterConeRadians = MathHelper.ToRadians(30.0f);
            light.Radius = LIGHT_RADIUS;

            // Initialize material settings.
            material.Ambient = new Color(new Vector4(0.2f, 0.2f, 0.2f, 1.0f));
            material.Diffuse = new Color(new Vector4(0.8f, 0.8f, 0.8f, 1.0f));
            material.Emissive = Color.Black;
            material.Specular = Color.Black;
            material.Shininess = 0.0f;

            // Parallax mapping height scale and bias values.
            scaleBias = new Vector2(0.04f, -0.03f);

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();
            currentMouseState = Mouse.GetState();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");

            effect = Content.Load<Effect>(@"Effects\parallax_normal_mapping");

            colorMap = Content.Load<Texture2D>(@"Textures\color_map");
            normalMap = Content.Load<Texture2D>(@"Textures\normal_map");
            heightMap = Content.Load<Texture2D>(@"Textures\height_map");

            // Create an empty white texture. This will be bound to the
            // colorMapTexture shader parameter when the user wants to
            // disable the color map texture. This trick will allow the
            // same shader to be used for when textures are enabled and
            // disabled.

            nullTexture = new Texture2D(GraphicsDevice, 1, 1, false, SurfaceFormat.Color);

            Color[] pixels = { Color.White };

            nullTexture.SetData(pixels);
        }

        protected override void UnloadContent()
        {
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (KeyJustPressed(Keys.Space))
            {
                if (++light.Type > Light.LightType.SpotLight)
                    light.Type = Light.LightType.DirectionalLight;
            }

            if (KeyJustPressed(Keys.Back))
            {
                if (--light.Type < Light.LightType.DirectionalLight)
                    light.Type = Light.LightType.SpotLight;
            }

            if (KeyJustPressed(Keys.B))
            {
                if (currentKeyboardState.IsKeyDown(Keys.LeftShift) ||
                    currentKeyboardState.IsKeyDown(Keys.RightShift))
                {
                    scaleBias.Y += 0.001f;
                }
                else
                {
                    scaleBias.Y -= 0.001f;
                }
            }

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (KeyJustPressed(Keys.T))
                disableColorMap = !disableColorMap;

            if (KeyJustPressed(Keys.P))
                disableParallax = !disableParallax;

            if (KeyJustPressed(Keys.S))
            {
                if (currentKeyboardState.IsKeyDown(Keys.LeftShift) ||
                    currentKeyboardState.IsKeyDown(Keys.RightShift))
                {
                    scaleBias.X += 0.01f;
                }
                else
                {
                    scaleBias.X -= 0.01f;
                }
            }

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) ||
                currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }
        }

        private void ProcessMouse()
        {
            prevMouseState = currentMouseState;
            currentMouseState = Mouse.GetState();

            float dx = 0.0f;
            float dy = 0.0f;
            float dz = 0.0f;

            if (currentMouseState.LeftButton == ButtonState.Pressed)
            {
                dx = currentMouseState.X - prevMouseState.X;
                dx *= CUBE_SPEED_TRACKING;

                dy = prevMouseState.Y - currentMouseState.Y;
                dy *= CUBE_SPEED_TRACKING;

                cubePos += Vector3.UnitX * dx;
                cubePos += Vector3.UnitY * dy;
            }
            else if (currentMouseState.MiddleButton == ButtonState.Pressed)
            {
                dz = currentMouseState.Y - prevMouseState.Y;
                dz *= CUBE_SPEED_DOLLYING;

                cubePos.Z += dz;

                if (cubePos.Z > CUBE_DOLLY_MAX)
                    cubePos.Z = CUBE_DOLLY_MAX;

                if (cubePos.Z < CUBE_DOLLY_MIN)
                    cubePos.Z = CUBE_DOLLY_MIN;
            }
            else if (currentMouseState.RightButton == ButtonState.Pressed)
            {
                dx = currentMouseState.X - prevMouseState.X;
                dx *= CUBE_SPEED_ROTATING;

                dy = currentMouseState.Y - prevMouseState.Y;
                dy *= CUBE_SPEED_ROTATING;

                cubeRotation.X += dy;
                cubeRotation.Y += dx;
            }
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();
        }

        protected override void Update(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            base.Update(gameTime);

            ProcessKeyboard();
            ProcessMouse();
            UpdateEffect();
            UpdateFrameRate(gameTime);
        }

        private void UpdateEffect()
        {
            // Calculate the perspective projection matrix.
            Matrix proj = Matrix.CreatePerspectiveFieldOfView(
                MathHelper.ToRadians(CAMERA_FOVY),
                (float)Window.ClientBounds.Width / (float)Window.ClientBounds.Height,
                CAMERA_ZNEAR, CAMERA_ZFAR);

            // Calculate the view matrix.
            Matrix view = Matrix.CreateLookAt(cameraPos, Vector3.Zero, Vector3.Up);

            // Calculate the world matrix.
            Matrix world = Matrix.CreateRotationY(MathHelper.ToRadians(cubeRotation.Y)) *
                           Matrix.CreateRotationX(MathHelper.ToRadians(cubeRotation.X)) *
                           Matrix.CreateTranslation(cubePos);

            // Calculate the transposed inverse of the world matrix.
            Matrix worldInvTrans = Matrix.Transpose(Matrix.Invert(world));

            // Calculate combined world-view-projection matrix.
            Matrix worldViewProj = world * view * proj;

            // Set shader matrix parameters.
            effect.Parameters["worldMatrix"].SetValue(world);
            effect.Parameters["worldInverseTransposeMatrix"].SetValue(worldInvTrans);
            effect.Parameters["worldViewProjectionMatrix"].SetValue(worldViewProj);

            // Set the shader camera position parameter.
            effect.Parameters["cameraPos"].SetValue(cameraPos);

            // Set the shader global ambiance parameters.
            effect.Parameters["globalAmbient"].SetValue(globalAmbient.ToVector4());

            // Set the shader parallax scale and bias parameter.
            effect.Parameters["scaleBias"].SetValue(scaleBias);

            // Set the shader lighting parameters.
            effect.Parameters["light"].StructureMembers["dir"].SetValue(light.Direction);
            effect.Parameters["light"].StructureMembers["pos"].SetValue(light.Position);
            effect.Parameters["light"].StructureMembers["ambient"].SetValue(light.Ambient.ToVector4());
            effect.Parameters["light"].StructureMembers["diffuse"].SetValue(light.Diffuse.ToVector4());
            effect.Parameters["light"].StructureMembers["specular"].SetValue(light.Specular.ToVector4());
            effect.Parameters["light"].StructureMembers["spotInnerCone"].SetValue(light.SpotInnerConeRadians);
            effect.Parameters["light"].StructureMembers["spotOuterCone"].SetValue(light.SpotOuterConeRadians);
            effect.Parameters["light"].StructureMembers["radius"].SetValue(light.Radius);

            // Set the shader material parameters.
            effect.Parameters["material"].StructureMembers["ambient"].SetValue(material.Ambient.ToVector4());
            effect.Parameters["material"].StructureMembers["diffuse"].SetValue(material.Diffuse.ToVector4());
            effect.Parameters["material"].StructureMembers["emissive"].SetValue(material.Emissive.ToVector4());
            effect.Parameters["material"].StructureMembers["specular"].SetValue(material.Specular.ToVector4());
            effect.Parameters["material"].StructureMembers["shininess"].SetValue(material.Shininess);

            // Bind the texture maps to the shader.
            effect.Parameters["colorMapTexture"].SetValue((disableColorMap) ? nullTexture : colorMap);
            effect.Parameters["normalMapTexture"].SetValue(normalMap);
            effect.Parameters["heightMapTexture"].SetValue(heightMap);

            // Select the shader based on light type.
            switch (light.Type)
            {
            case Light.LightType.DirectionalLight:
                if (disableParallax)
                    effect.CurrentTechnique = effect.Techniques["NormalMappingDirectionalLighting"];
                else
                    effect.CurrentTechnique = effect.Techniques["ParallaxNormalMappingDirectionalLighting"];
                break;

            case Light.LightType.PointLight:
                if (disableParallax)
                    effect.CurrentTechnique = effect.Techniques["NormalMappingPointLighting"];
                else
                    effect.CurrentTechnique = effect.Techniques["ParallaxNormalMappingPointLighting"];
                break;

            case Light.LightType.SpotLight:
                if (disableParallax)
                    effect.CurrentTechnique = effect.Techniques["NormalMappingSpotLighting"];
                else
                    effect.CurrentTechnique = effect.Techniques["ParallaxNormalMappingSpotLighting"];
                break;

            default:
                break;
            }
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        private void DrawText()
        {
            StringBuilder buffer = new StringBuilder();

            if (displayHelp)
            {
                buffer.AppendLine("Left mouse click and drag to track cube");
                buffer.AppendLine("Middle mouse click and drag to dolly cube");
                buffer.AppendLine("Right mouse click and drag to rotate cube");
                buffer.AppendLine();
                buffer.AppendLine("Press B to increase the parallax bias.");
                buffer.AppendLine("Press b to decrease the parallax bias.");
                buffer.AppendLine("Press P to enable/disable parallax");
                buffer.AppendLine("Press S to increase the parallax scale.");
                buffer.AppendLine("Press s to decrease the parallax scale.");
                buffer.AppendLine("Press T to enable/disable textures");
                buffer.AppendLine("Press SPACE to cycle forwards through the light types");
                buffer.AppendLine("Press BACKSPACE to cycle backwards through the light types");
                buffer.AppendLine("Press ALT + ENTER to toggle full screen");
                buffer.AppendLine("Press ESCAPE to exit");
                buffer.AppendLine();
                buffer.AppendLine("Press H to hide help");
            }
            else
            {
                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                buffer.AppendFormat("Lighting: {0}\n", light.Type);

                if (disableParallax)
                {
                    buffer.AppendLine("Technique: Normal Mapping");
                }
                else
                {
                    buffer.AppendLine("Technique: Parallax Normal Mapping");
                    buffer.AppendFormat("  Parallax scale: {0}\n", scaleBias.X.ToString("f3"));
                    buffer.AppendFormat("  Parallax bias: {0}\n", scaleBias.Y.ToString("f3"));
                }

                buffer.AppendLine();
                buffer.AppendLine("Press H to display help");
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, buffer.ToString(), fontPos, Color.Yellow);
            spriteBatch.End();
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!this.IsActive)
                return;

            GraphicsDevice.Clear(Color.CornflowerBlue);

            cube.Draw(effect, GraphicsDevice);
            DrawText();

            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}